"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const etag_1 = __importDefault(require("etag"));
const response_end_overrider_1 = require("./utils/response-end-overrider");
class ExpressCache {
    constructor(opts) {
        this.readCacheErrHandler = (err) => {
            throw err;
        };
        this.storeCacheErrHandler = (err) => {
            throw err;
        };
        this.clearCacheErrHandler = (err) => {
            throw err;
        };
        if (!opts.redisClient) {
            throw new Error('redis client is required');
        }
        this.keyPrefix = opts.keyPrefix || 'expch';
        this.ttl = opts.ttl || 60;
        this.redisClient = opts.redisClient;
        this.requestHeaderName = opts.requestHeaderName || 'if-none-match';
        this.responseHeaderName = opts.responseHeaderName || 'etag';
        this.tagPrefix = '';
        if (opts.readCacheErrHandler && typeof opts.readCacheErrHandler === 'function') {
            this.readCacheErrHandler = opts.readCacheErrHandler;
        }
        if (opts.storeCacheErrHandler && typeof opts.storeCacheErrHandler === 'function') {
            this.storeCacheErrHandler = opts.storeCacheErrHandler;
        }
        if (opts.clearCacheErrHandler && typeof opts.clearCacheErrHandler === 'function') {
            this.clearCacheErrHandler = opts.clearCacheErrHandler;
        }
    }
    middleware(opts) {
        let isCanSaveFn = (httpStatusCode) => httpStatusCode >= 200 && httpStatusCode <= 299;
        if (opts.isCanSaveFn && typeof opts.isCanSaveFn === 'function') {
            isCanSaveFn = opts.isCanSaveFn;
        }
        return (req, res, next) => __awaiter(this, void 0, void 0, function* () {
            if (opts.tagPrefix && typeof opts.tagPrefix === 'string') {
                this.tagPrefix = opts.tagPrefix;
            }
            else if (opts.tagPrefix && typeof opts.tagPrefix === 'function') {
                this.tagPrefix = opts.tagPrefix(req);
            }
            else {
                const queryString = Object.keys(req.query).length > 0 ? JSON.stringify(req.query) : '';
                this.tagPrefix = req.originalUrl + queryString;
            }
            // response from cache
            try {
                const tag = req.header(this.requestHeaderName);
                if (tag) {
                    const key = this.getKey(tag);
                    const dataCache = yield this.getCache(key);
                    if (dataCache) {
                        res.status(dataCache.httpStatusCode).set(dataCache.headers).end(dataCache.body);
                        return;
                    }
                }
            }
            catch (err) {
                yield this.readCacheErrHandler(err, req, res, next);
                return;
            }
            // store response to cache
            const fn = (resBody) => __awaiter(this, void 0, void 0, function* () {
                const isAllowSaveCache = yield isCanSaveFn(res.statusCode, res, req);
                if (!isAllowSaveCache) {
                    return;
                }
                try {
                    const tag = (0, etag_1.default)(resBody, { weak: false });
                    res.header(this.responseHeaderName, tag);
                    const data = {
                        httpStatusCode: res.statusCode,
                        headers: res.getHeaders(),
                        body: resBody,
                    };
                    const key = this.getKey(tag);
                    yield this.setCache(key, data);
                }
                catch (err) {
                    yield this.storeCacheErrHandler(err, req, res, next);
                }
            });
            (0, response_end_overrider_1.responseEndOverrider)(res, fn);
            next();
        });
    }
    getKey(tag) {
        return this.keyPrefix + ':' + this.tagPrefix + tag;
    }
    getCache(key) {
        return __awaiter(this, void 0, void 0, function* () {
            const data = (yield this.redisClient.get(key));
            if (!data) {
                return null;
            }
            const cache = JSON.parse(data);
            return cache;
        });
    }
    setCache(key, data) {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.redisClient.set(key, JSON.stringify(data), {
                EX: this.ttl,
            });
        });
    }
    clear(opts) {
        const keyPatterns = [];
        let isCanClearFn = (httpStatusCode) => httpStatusCode >= 200 && httpStatusCode <= 299;
        if (opts.isCanClearFn && typeof opts.isCanClearFn === 'function') {
            isCanClearFn = opts.isCanClearFn;
        }
        return (req, res, next) => __awaiter(this, void 0, void 0, function* () {
            if (opts.tagPattern && typeof opts.tagPattern === 'function') {
                opts.tagPattern = opts.tagPattern(req);
            }
            if (opts.tagPattern && typeof opts.tagPattern === 'string') {
                keyPatterns.push(this.getKey(opts.tagPattern));
            }
            else if (Array.isArray(opts.tagPattern) && opts.tagPattern.length > 0) {
                for (const tp of opts.tagPattern) {
                    keyPatterns.push(this.getKey(tp));
                }
            }
            if (keyPatterns.length > 0) {
                const fn = () => __awaiter(this, void 0, void 0, function* () {
                    const isCanClear = yield isCanClearFn(res.statusCode, res, req);
                    if (!isCanClear) {
                        return;
                    }
                    try {
                        const results = yield Promise.all(keyPatterns.map((keyPattern) => this.redisClient.keys(keyPattern)));
                        const keys = [].concat(...results);
                        if (keys.length > 0) {
                            yield this.redisClient.del(keys);
                        }
                    }
                    catch (err) {
                        yield this.clearCacheErrHandler(err, req, res, next);
                    }
                });
                (0, response_end_overrider_1.responseEndOverrider)(res, fn);
            }
            next();
        });
    }
}
exports.default = ExpressCache;
